# -*- coding: utf-8 -*-

import hashlib
from .encoding import decode_root


def bit_is_set(bits, i):
    return bits[int(i/8)] & (1 << (7-i % 8)) != 0


def verify_inclusion(ap, root, key, value):
    """ verify_inclusion verifies the merkle proof 'ap' (audit path) that
    the key/value pair in included in the trie with root 'root'.
    """
    leaf_hash = hashlib.sha256(key + value + bytes([256-len(ap)])).digest()
    return root == verify_proof(ap, 0, key, leaf_hash)


def verify_inclusion_c(ap, height, bitmap, root, key, value):
    """ verify_inclusion verifies the compressed merkle proof 'ap' (audit path)
    that the key/value pair in included in the trie with root 'root'.
    """
    leaf_hash = hashlib.sha256(key + value + bytes([256-height])).digest()
    return root == verify_proof_c(bitmap, key, leaf_hash, ap, height, 0, 0)


def verify_proof(ap, key_index, key, leaf_hash):
    """ verify_proof recursively hashes the result with the proof nodes in the
    audit path 'ap'
    """
    if key_index == len(ap):
        return leaf_hash

    if bit_is_set(key, key_index):
        right = verify_proof(ap, key_index+1, key, leaf_hash)
        return hashlib.sha256(ap[len(ap)-key_index-1] + right).digest()
    left = verify_proof(ap, key_index+1, key, leaf_hash)
    return hashlib.sha256(left + ap[len(ap)-key_index-1]).digest()


def verify_proof_c(bitmap, key, leaf_hash, ap, length, key_index, ap_index):
    """ verify_proof_c recursively hashes the result with the proof nodes in
    the compressed audit path 'ap'
    """
    if key_index == length:
        return leaf_hash

    if bit_is_set(key, key_index):
        if bit_is_set(bitmap, length-key_index-1):
            right = verify_proof_c(bitmap, key, leaf_hash, ap, length,
                                   key_index+1, ap_index+1)
            return hashlib.sha256(ap[len(ap)-ap_index-1] + right).digest()
        left = verify_proof_c(bitmap, key, leaf_hash, ap, length,
                              key_index+1, ap_index)
        return hashlib.sha256(bytes([0]) + left).digest()

    if bit_is_set(bitmap, length-key_index-1):
        right = verify_proof_c(bitmap, key, leaf_hash, ap, length,
                               key_index+1, ap_index+1)
        return hashlib.sha256(right + ap[len(ap)-ap_index-1]).digest()
    left = verify_proof_c(bitmap, key, leaf_hash, ap, length,
                          key_index+1, ap_index)
    return hashlib.sha256(left + bytes([0])).digest()


def verify_exclusion(root, ap, key, proofKey, proofVal):
    """ verify_exclusion verifies the merkle proof that a default
    node (bytes([0]) is included on the path of the 'key', or that the
    proofKey/proofVal key pair is included on the path of the 'key'
    """
    if isinstance(root, str) and len(root) != 0:
        root = decode_root(root)

    if not proofKey:
        # return true if a DefaultLeaf in the key path is included in the trie
        return root == verify_proof(ap, 0, key, bytes([0]))

    # Check if another kv leaf is on the key path in 2 steps
    # 1- Check the proof leaf exists
    if not verify_inclusion(ap, root, proofKey, proofVal):
        # the proof leaf is not included in the trie
        return False

    # 2- Check the proof leaf is on the key path
    for b in range(len(ap)):
        if bit_is_set(key, b) != bit_is_set(proofKey, b):
            # the proofKey leaf node is not on the path of the key
            return False
    # return true because we verified proofKey/proofVal is on the key path
    return True


def verify_exclusion_c(root, ap, length, bitmap, key, proofKey, proofVal):
    """ verify_exclusion_c verifies the compressed merkle proof that a default
    node (bytes([0]) is included on the path of the 'key', or that the
    proofKey/proofVal key pair is included on the path of the 'key'
    """
    if isinstance(root, str) and len(root) != 0:
        root = decode_root(root)

    if not proofKey:
        # return true if a DefaultLeaf in the key path is included in the trie
        return root == verify_proof_c(bitmap, key, bytes([0]), ap,
                                      length, 0, 0)

    # Check if another kv leaf is on the key path in 2 steps
    # 1- Check the proof leaf exists
    if not verify_inclusion_c(ap, length, bitmap, root, proofKey, proofVal):
        # the proof leaf is not included in the trie
        return False

    # 2- Check the proof leaf is on the key path
    for b in range(length):
        if bit_is_set(key, b) != bit_is_set(proofKey, b):
            # the proofKey leaf node is not on the path of the key
            return False
    # return true because we verified proofKey/proofVal is on the key path
    return True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: polarrpc.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import node_pb2 as node__pb2
from . import rpc_pb2 as rpc__pb2
from . import metric_pb2 as metric__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='polarrpc.proto',
  package='types',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x0epolarrpc.proto\x12\x05types\x1a\nnode.proto\x1a\trpc.proto\x1a\x0cmetric.proto\"(\n\x0bPaginations\x12\x0b\n\x03ref\x18\x01 \x01(\x0c\x12\x0c\n\x04size\x18\x03 \x01(\r\"T\n\x0fPolarisPeerList\x12\r\n\x05total\x18\x01 \x01(\r\x12\x0f\n\x07hasNext\x18\x02 \x01(\x08\x12!\n\x05peers\x18\x03 \x03(\x0b\x32\x12.types.PolarisPeer\"h\n\x0bPolarisPeer\x12#\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0b\x32\x12.types.PeerAddress\x12\x11\n\tconnected\x18\x02 \x01(\x03\x12\x11\n\tlastCheck\x18\x03 \x01(\x03\x12\x0e\n\x06verion\x18\x04 \x01(\t2\xac\x02\n\x11PolarisRPCService\x12\x31\n\tNodeState\x12\x0e.types.NodeReq\x1a\x12.types.SingleBytes\"\x00\x12\x31\n\x06Metric\x12\x15.types.MetricsRequest\x1a\x0e.types.Metrics\"\x00\x12;\n\x0b\x43urrentList\x12\x12.types.Paginations\x1a\x16.types.PolarisPeerList\"\x00\x12\x39\n\tWhiteList\x12\x12.types.Paginations\x1a\x16.types.PolarisPeerList\"\x00\x12\x39\n\tBlackList\x12\x12.types.Paginations\x1a\x16.types.PolarisPeerList\"\x00\x62\x06proto3')
  ,
  dependencies=[node__pb2.DESCRIPTOR,rpc__pb2.DESCRIPTOR,metric__pb2.DESCRIPTOR,])




_PAGINATIONS = _descriptor.Descriptor(
  name='Paginations',
  full_name='types.Paginations',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ref', full_name='types.Paginations.ref', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='types.Paginations.size', index=1,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=62,
  serialized_end=102,
)


_POLARISPEERLIST = _descriptor.Descriptor(
  name='PolarisPeerList',
  full_name='types.PolarisPeerList',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='total', full_name='types.PolarisPeerList.total', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hasNext', full_name='types.PolarisPeerList.hasNext', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='peers', full_name='types.PolarisPeerList.peers', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=104,
  serialized_end=188,
)


_POLARISPEER = _descriptor.Descriptor(
  name='PolarisPeer',
  full_name='types.PolarisPeer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='types.PolarisPeer.address', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='connected', full_name='types.PolarisPeer.connected', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lastCheck', full_name='types.PolarisPeer.lastCheck', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='verion', full_name='types.PolarisPeer.verion', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=294,
)

_POLARISPEERLIST.fields_by_name['peers'].message_type = _POLARISPEER
_POLARISPEER.fields_by_name['address'].message_type = node__pb2._PEERADDRESS
DESCRIPTOR.message_types_by_name['Paginations'] = _PAGINATIONS
DESCRIPTOR.message_types_by_name['PolarisPeerList'] = _POLARISPEERLIST
DESCRIPTOR.message_types_by_name['PolarisPeer'] = _POLARISPEER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Paginations = _reflection.GeneratedProtocolMessageType('Paginations', (_message.Message,), dict(
  DESCRIPTOR = _PAGINATIONS,
  __module__ = 'polarrpc_pb2'
  # @@protoc_insertion_point(class_scope:types.Paginations)
  ))
_sym_db.RegisterMessage(Paginations)

PolarisPeerList = _reflection.GeneratedProtocolMessageType('PolarisPeerList', (_message.Message,), dict(
  DESCRIPTOR = _POLARISPEERLIST,
  __module__ = 'polarrpc_pb2'
  # @@protoc_insertion_point(class_scope:types.PolarisPeerList)
  ))
_sym_db.RegisterMessage(PolarisPeerList)

PolarisPeer = _reflection.GeneratedProtocolMessageType('PolarisPeer', (_message.Message,), dict(
  DESCRIPTOR = _POLARISPEER,
  __module__ = 'polarrpc_pb2'
  # @@protoc_insertion_point(class_scope:types.PolarisPeer)
  ))
_sym_db.RegisterMessage(PolarisPeer)



_POLARISRPCSERVICE = _descriptor.ServiceDescriptor(
  name='PolarisRPCService',
  full_name='types.PolarisRPCService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=297,
  serialized_end=597,
  methods=[
  _descriptor.MethodDescriptor(
    name='NodeState',
    full_name='types.PolarisRPCService.NodeState',
    index=0,
    containing_service=None,
    input_type=rpc__pb2._NODEREQ,
    output_type=rpc__pb2._SINGLEBYTES,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='Metric',
    full_name='types.PolarisRPCService.Metric',
    index=1,
    containing_service=None,
    input_type=metric__pb2._METRICSREQUEST,
    output_type=metric__pb2._METRICS,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='CurrentList',
    full_name='types.PolarisRPCService.CurrentList',
    index=2,
    containing_service=None,
    input_type=_PAGINATIONS,
    output_type=_POLARISPEERLIST,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='WhiteList',
    full_name='types.PolarisRPCService.WhiteList',
    index=3,
    containing_service=None,
    input_type=_PAGINATIONS,
    output_type=_POLARISPEERLIST,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='BlackList',
    full_name='types.PolarisRPCService.BlackList',
    index=4,
    containing_service=None,
    input_type=_PAGINATIONS,
    output_type=_POLARISPEERLIST,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_POLARISRPCSERVICE)

DESCRIPTOR.services_by_name['PolarisRPCService'] = _POLARISRPCSERVICE

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

import json

from .address import Address
from .block_hash import BlockHash
from .tx_hash import TxHash


class Event:
    def __init__(self, grpc_event):
        self.__contract_address = Address.encode(grpc_event.contractAddress)
        self.__name = grpc_event.eventName
        self.__index = grpc_event.eventIdx
        try:
            self.__args = json.loads(grpc_event.jsonArgs)
        except:
            self.__args = [grpc_event.jsonArgs]
        self.__block_hash = BlockHash(grpc_event.blockHash)
        self.__block_height = grpc_event.blockNo
        self.__tx_hash = TxHash(grpc_event.txHash)
        self.__tx_index = grpc_event.txIndex

    @property
    def tx_hash(self):
        return self.__tx_hash

    @property
    def tx_index(self):
        return self.__tx_index

    @property
    def block_hash(self):
        return self.__block_hash

    @property
    def block_height(self):
        return self.__block_height

    @property
    def contract_address(self):
        return self.__contract_address

    @property
    def index(self):
        return self.__index

    @property
    def name(self):
        return self.__name

    @property
    def arguments(self):
        return self.__args

    def json(self):
        return {
            'contract_address': self.contract_address,
            'event_name': self.name,
            'index': self.index,
            'arguments': self.arguments,
            'block_hash': str(self.block_hash),
            'block_height': self.block_height,
            'tx_hash': str(self.tx_hash),
            'tx_index': self.tx_index,
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)



# -*- coding: utf-8 -*-

import json

from google.protobuf.json_format import MessageToJson

from .address import Address

class NameInfo():
    """ NameInfo is used to store information of name system."""
    def __init__(self, info):
        self.__info = json.loads(MessageToJson(info))
        self.__name = info.name.name
        self.__owner = Address.encode(info.owner)
        self.__dest = Address.encode(info.destination)

    @property
    def info(self):
        return self.__info

    @property
    def name(self):
        return self.__name

    @property
    def owner(self):
        return self.__owner

    @property
    def destination(self):
        return self.__dest

    def json(self):
        return {
            'name': self.name,
            'owner': self.owner,
            'destination': self.destination
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)



import sys
import traceback
import json

import aergo.herapy as herapy

from aergo.herapy.utils.converter import convert_bytes_to_hex_str, \
                                         convert_bytes_to_int_str, get_hash


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Get Block Meta Stream -----------")
        stream = aergo.receive_block_stream()
        i = 0
        while i < 3:
            block = next(stream)
            print('[{}] block: {}'.format(i, str(block)))
            i += 1

        print("------ Get Block Stream -----------")
        stream = aergo.receive_block_stream()
        i = 0
        while i < 3:
            block = next(stream)
            print('[{}] block: {}'.format(i, str(block)))
            i += 1

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



import pytest

from ecdsa.ecdsa import Public_key

from aergo.herapy.obj.private_key import PrivateKey
from aergo.herapy.utils.encoding import decode_private_key

PK_STR = "6jEuQyF6RZ9xpk7s99LzHoY2X7MeT8V88HCrsewwmynbKXtwUxZ"


def test_fail():
    with pytest.raises(TypeError):
        PrivateKey()

    with pytest.raises(TypeError):
        PrivateKey(1234)

    with pytest.raises(ValueError):
        PrivateKey("1234")


def test_success():
    # generate new private key
    pk = PrivateKey(None)
    assert type(pk) is PrivateKey
    assert isinstance(str(pk), str)
    assert isinstance(bytes(pk), bytes)
    assert type(pk.public_key) is Public_key

    # from private string
    pk = PrivateKey(PK_STR)
    assert type(pk) is PrivateKey
    assert isinstance(str(pk), str)
    assert PK_STR == str(pk)
    assert isinstance(bytes(pk), bytes)
    assert decode_private_key(PK_STR) == bytes(pk)
    assert type(pk.public_key) is Public_key

    # from private bytes
    pk = PrivateKey(decode_private_key(PK_STR))
    assert type(pk) is PrivateKey
    assert isinstance(str(pk), str)
    assert PK_STR == str(pk)
    assert isinstance(bytes(pk), bytes)
    assert decode_private_key(PK_STR) == bytes(pk)
    assert type(pk.public_key) is Public_key


def test_sign():
    msg = "test"
    msg_bytes = msg.encode('utf-8')

    pk = PrivateKey(None)
    with pytest.raises(TypeError):
        pk.sign_msg(msg)
    sign = pk.sign_msg(msg_bytes)
    assert isinstance(sign, bytes)
    with pytest.raises(TypeError):
        pk.verify_sign(msg, sign)
    assert pk.verify_sign(msg_bytes, sign)
    assert not pk.verify_sign("test1".encode('utf-8'), sign)
    with pytest.raises(TypeError):
        pk.verify_sign(msg_bytes, b'\x10')
    with pytest.raises(IndexError):
        pk.verify_sign(msg_bytes, b'\x30')
    with pytest.raises(TypeError):
        pk.verify_sign(msg_bytes, b'\x30\x02')
    with pytest.raises(TypeError):
        pk.verify_sign(msg_bytes, b'\x30\x04\x00\x00\x00\x00')
    with pytest.raises(TypeError):
        pk.verify_sign(msg_bytes, b'\x30\x04\x02\x00\x00\x00')
    with pytest.raises(ValueError):
        pk.verify_sign(msg_bytes, b'\x30\x04\x02\x00\x02\x00')


def test_asym_sec_msg():
    msg = "asymmetric encrypt/decrypt test"

    pk1 = PrivateKey(None)
    pk2 = PrivateKey(None)

    enc_msg = pk1.asymmetric_encrypt_msg(pk2.address, msg)
    dec_msg = pk2.asymmetric_decrypt_msg(pk1.address, enc_msg)
    dec_msg = str(dec_msg, encoding='utf-8')

    assert msg == dec_msg
